import asyncio
import os
import sys
from locks import Mutex

from ytrss.configuration.entity.destination_info import DestinationId
from ytrss.core.entity.destination import Destination
from ytrss.core.entity.downloader import DownloaderError
from ytrss.core.entity.movie import Movie
from ytrss.core.helpers.logging import logger
from ytrss.core.managers.manager_service import ManagerService, default_manager_service
from ytrss.database.database import DatabaseStatus


async def download_movie(
        movie: Movie,
        destination: Destination,
        manager_service: ManagerService = default_manager_service()
) -> None:
    """ Download movie

    This function download movie and save in destination, but it not check any conditions.
    """
    downloaded_movie = manager_service.plugin_manager.create_downloader(movie, manager_service.configuration).download()
    os.makedirs('/tmp/ytrss', exist_ok=True)
    with Mutex(f'/tmp/ytrss/destination.{destination.identity}.lock', timeout=5.0):
        destination.save(downloaded_movie.data_paths)


async def download_task(
        movie: Movie,
        destination_id: DestinationId,
        manager_service: ManagerService = default_manager_service()
) -> bool:
    """ Download movie

    The function make a lock on movie, download it and send to destination. It also checks if file
    can be downloaded and return with False value if not.
    """
    os.makedirs('/tmp/ytrss', exist_ok=True)
    try:
        with Mutex(f'/tmp/ytrss/movie-{movie.identity}.lock'):
            if not manager_service.database.is_new(movie, destination_id):
                return False
            if not movie.is_ready:
                logger.warning("Movie is not ready to download [%s] %s",
                               movie.url,
                               movie.title)
                manager_service.database.change_type(movie, DatabaseStatus.WAIT)
                return False
            destination = manager_service.destination_manager[destination_id]
            try:
                manager_service.database.change_type(movie, DatabaseStatus.PROGRESS)
                await download_movie(movie, destination)
                logger.info("Movie downloaded: [%s] %s",
                            movie.url,
                            movie.title)
                manager_service.database.change_type(movie, DatabaseStatus.DONE)
                return True
            except DownloaderError:
                logger.error("Cannot download movie: [%s] %s",
                             movie.url,
                             movie.title)
                manager_service.database.change_type(movie, DatabaseStatus.ERROR)
    except BlockingIOError:
        logger.error("Cannot download movie: [%s] %s",
                     movie.url,
                     movie.title)
    return False


def download_all_movies(manager_service: ManagerService = default_manager_service()) -> int:
    """ Download all movies.

    This function download all files from database and send it to destination place. All the
    movies are download as a separate async task.
    """

    logger.info("Starting download movies:")
    try:
        loop = asyncio.get_event_loop()
        outputs = loop.run_until_complete(asyncio.gather(
            *[download_task(movie, destination) for movie, destination in
              manager_service.database.movies()]
        ))
        queue_len = len(outputs)
        downloaded = len(list(filter(lambda x: x, outputs)))
        loop.close()

    except KeyboardInterrupt:
        logger.info("Keyboard Interrupt by user.")
        sys.exit(1)
    except Exception as ex:
        logger.error("Unexpected Error: %s", type(ex))
        raise ex
    if queue_len == 0:
        logger.warning("Cannot find url to download")
    elif queue_len >= 0 and downloaded < queue_len:
        logger.warning("Cannot download all movies %s/%s", downloaded, queue_len)
    else:
        logger.info("Download all %s movie", downloaded)
    return downloaded



import copy
from json import JSONDecodeError
from typing import Optional

import sys
import json
from datetime import datetime
from io import StringIO
import youtube_dl

from ytrss.core.entity.movie import Movie, MovieError
from ytrss.core.helpers.typing import Url


class InvalidDataMovieError(MovieError):
    """ Invalid parameter movie error

    This exception is invoked when the movie data is invalid
    """


class YouTubeMovie(Movie):
    """ A YouTube implementation of movie """

    def __init__(self, url: Url) -> None:
        self._url: Url = url
        self._date: Optional[datetime] = None

        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = tmp_stdout = StringIO()
        sys.stderr = tmp_stderr = StringIO()
        try:
            youtube_dl.main(['--dump-json', url])
        except SystemExit:
            pass
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        json_output = tmp_stdout.getvalue()
        self._error = tmp_stderr.getvalue()
        self._json_data = {}
        try:
            self._json_data = json.load(StringIO(json_output))
        except JSONDecodeError:
            raise InvalidDataMovieError(f"Invalid address type: [{url}] -> {self._error}")

    @property
    def url(self) -> Url:
        return self._url

    @property
    def identity(self) -> str:
        extractor: str = self.__get_youtube_data("extractor")
        code: str = self.__get_youtube_data("display_id")
        return f"ytdl:{extractor}:{code}"

    def __get_youtube_data(self, key: str) -> str:
        return str(self._json_data.get(key, ""))

    @property
    def title(self) -> str:
        return self.__get_youtube_data("title")

    @property
    def author(self) -> str:
        return self.__get_youtube_data('uploader')

    @property
    def description(self) -> str:
        return self.__get_youtube_data("description")

    @property
    def date(self) -> datetime:
        try:
            return datetime.strptime(self.__get_youtube_data("upload_date"), "%Y%m%d")
        except ValueError:
            return datetime.now()

    @property
    def img_url(self) -> Optional[Url]:
        img = self.__get_youtube_data("thumbnail")
        return Url(img) if img != "" else None

    @property
    def is_ready(self) -> bool:
        is_live = "false"
        if self._error is None:
            is_live = self.__get_youtube_data("is_live")

        return ("Premieres in" not in self._error if self._error is not None else True) and is_live.lower() != "true"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Movie):
            return self.identity == other.identity
        if other == "":
            return False
        tmp_other = copy.deepcopy(other)
        return tmp_other == self

    def __str__(self) -> str:
        return self.url



from typing import Dict, Iterator

from ytrss.configuration.entity.destination_info import DestinationId, DestinationInfo
from ytrss.core.entity.destination import Destination, DestinationError
from ytrss.core.helpers.logging import logger
from ytrss.core.managers.plugin_manager import PluginManager


class DestinationManager:
    """
    Destination manager

    An object that managed all destinations available in application.
    """

    def __init__(self, plugin_manager: PluginManager) -> None:
        self._plugin_manager = plugin_manager
        self._destinations: Dict[DestinationId, Destination] = {}

    def add_from_info(self, info: DestinationInfo) -> None:
        """
        Add a destination from DestinationInfo object.
        """
        try:
            self._destinations[info.identity] = self._plugin_manager.create_destination(info)
        except DestinationError:
            logger.error("Cannot create destination from info: %s (type=%s)", info, info.type)

    def __contains__(self, item: DestinationId) -> bool:
        return item in self._destinations

    def __getitem__(self, key: DestinationId) -> Destination:
        if key in self._destinations:
            return self._destinations[key]
        raise KeyError  # TODO: Default destination from code

    def __iter__(self) -> Iterator[Destination]:
        for key in self._destinations:
            yield self._destinations[key]

    @property
    def destinations(self) -> Iterator[Destination]:
        """
        Creator of all destinations available from manager.
        """
        for key in self._destinations:
            yield self._destinations[key]



import os
import json

from typing import Any

from ytrss.configuration.configuration import ConfigurationError, ConfigurationFileNotExistsError
from ytrss.configuration.readers.base import ConfigurationReader


class JsonConfigurationParseError(ConfigurationError):
    """ Json configuration error. """


class JsonConfigurationFileNotExistsError(ConfigurationFileNotExistsError):
    """ Json configuration file not exists """


class JsonConfiguration(ConfigurationReader):
    """ Configuration json object """

    @property
    def read_config(self) -> Any:
        return self.__configuration

    def __init__(self, conf_file: str) -> None:
        if not os.path.isfile(os.path.expanduser(conf_file)):
            raise ConfigurationError("File not exists")
        try:
            with open(os.path.expanduser(conf_file)) as data_file:
                data = json.load(data_file)
        except FileNotFoundError:
            raise JsonConfigurationFileNotExistsError()

        if data == {}:
            raise JsonConfigurationParseError("Cannot find data from file")

        self.__configuration = data



import os
from contextlib import contextmanager

from typing import Iterator


@contextmanager
def cwd(cwd_path: str) -> Iterator[None]:
    """ Change work directory

    This context manager change work directory where the script was executed. It's also back to
    previous directory, at the end of script.

    .. warning::
        After end of 'with' structure, the work directory will be back to previous state, even
        if it's change according the structure.
    """
    old_path = os.getcwd()
    os.chdir(cwd_path)
    yield
    os.chdir(old_path)



from typing import Dict, Any

from ytrss.configuration.readers.base import ConfigurationReader


class DefaultConfiguration(ConfigurationReader):
    """ Default configuration """

    @property
    def read_config(self) -> Dict[str, Any]:
        return {}



import abc
from datetime import datetime
from typing import Optional

from ytrss.core.helpers.typing import Url


class MovieError(Exception):
    """ Movie error

    An error raised when the movie cannot be created or there are some issue with creation.
    """


class Movie(metaclass=abc.ABCMeta):
    """ A movie abstract class

    This class represents a movie data. It should have an information about it, and be serialized to
    and from url.
    """

    @property
    @abc.abstractmethod
    def identity(self) -> str:
        """ A unique identity of the movie

        This identity is used to check if the two movies are the same.
        """

    @property
    @abc.abstractmethod
    def url(self) -> Url:
        """ Movie's url """

    @property
    @abc.abstractmethod
    def title(self) -> str:
        """ Movie's title """

    @property
    @abc.abstractmethod
    def author(self) -> str:
        """ Movie's author """

    @property
    @abc.abstractmethod
    def description(self) -> str:
        """ Movie's description """

    @property
    @abc.abstractmethod
    def date(self) -> datetime:
        """ Create data """

    @property
    @abc.abstractmethod
    def img_url(self) -> Optional[Url]:
        """ An Url to movie's image """

    @property
    @abc.abstractmethod
    def is_ready(self) -> bool:
        """ Tell if movie is ready to download """
        return True



import abc

from enum import Enum

from typing import Iterable, Tuple

from ytrss.configuration.entity.destination_info import DestinationId
from ytrss.core.entity.movie import Movie


class DatabaseStatus(Enum):
    """
    Database status types
    """
    WAIT = "wait"
    PROGRESS = "progress"
    DONE = "done"
    ERROR = "error"


class Database(metaclass=abc.ABCMeta):
    """
    Abstract database class
    """

    @abc.abstractmethod
    def is_new(self, movie: Movie, destination: DestinationId) -> bool:
        """ Check if movie is ready to download. """

    @abc.abstractmethod
    def change_type(self, movie: Movie, db_type: DatabaseStatus) -> None:
        """ Change type of movie task. """

    @abc.abstractmethod
    def movies(self) -> Iterable[Tuple[Movie, DestinationId]]:
        """ Get list of movies to download. """

    @abc.abstractmethod
    def queue_mp3(self, movie: Movie, destination: DestinationId) -> bool:
        """ Add movie task to queue. """



import re
import time
from datetime import datetime
from email import utils

from typing import Optional


def format_desc(text: Optional[str]) -> str:
    """
    Format description.

    This function replace ulr and newline chars by following html code.
    It is using in rss file to format description in rss reader.
    """
    if text is None:
        return ""
    result = text
    result = re.sub(r"[\w\.\+\-]+\@[\w]+\.[a-z]{2,3}",
                    r"<a href='mailto:\g<0>'>\g<0></a>",
                    result)
    result = re.sub(r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|i"
                    "(?:%[0-9a-fA-F][0-9a-fA-F]))+",
                    r"<a href='\g<0>'>\g<0></a>",
                    result)
    result = re.sub(r"\/redirect(?:[a-zA-Z0-9$-_@.&!*\(\)]*)(?:q=(?P<url>[A-Za-z0-9%.]+))+"
                    "(?:[a-zA-Z0-9$-_@.&!*,]*)",
                    r"<a href='\g<url>'>\g<url></a>",
                    result)
    result = re.sub(r"\/(?:redirect|watch)(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|"
                    "(?:%[0-9a-zA-Z][0-9a-zA-Z]))+",
                    r"<a href='https://youtube.com\g<0>'>\g<0></a>",
                    result)
    result = result.replace("\n", "<br />\n")
    result = result.replace("%3A", ":")
    result = result.replace("%2F", "/")
    return result


def format_str(text: Optional[str]) -> str:
    """
    Format some strings.

    This function replacing invalid chars in xml file.
    ex. RSS doesn't accept C{&} char, so it should be replaced to C{&amp;}
    """
    return text.replace("&", "&amp;") if text is not None else ""


def format_date(date: Optional[datetime]) -> str:
    """
    Convert datetime to string
    """
    print_date: datetime = date if date is not None else datetime.now()
    return utils.formatdate(time.mktime(print_date.timetuple()))



import os
import sys

from typing import Dict, Sequence, Iterable

from ytrss.configuration.configuration import Configuration
from ytrss.configuration.entity.destination_info import DestinationId, DestinationInfo
from ytrss.configuration.entity.source import Source
from ytrss.core.helpers.typing import Path


class YtrssConfiguration:
    """ Configuration data. """

    # pylint: disable=R0912,R0915
    def __init__(self, configuration: Configuration) -> None:
        """
        Configuration data constructor.
        """
        self._configuration_data = configuration.conf

    @property
    def destinations(self) -> Dict[str, DestinationInfo]:
        """ Destinations of destinations """
        result: Dict[str, DestinationInfo] = {}
        if 'destinations' in self._configuration_data and isinstance(self._configuration_data['destinations'], dict):
            for key, value in self._configuration_data['destinations'].items():
                result[key] = DestinationInfo.from_json(value, DestinationId(key))
        return result

    @property
    def sources(self) -> Iterable[Source]:
        """ List of sources """
        if 'subscriptions' in self._configuration_data and isinstance(self._configuration_data['subscriptions'], list):
            for subscription in self._configuration_data['subscriptions']:
                source = Source.from_json(subscription)
                if source.enable:
                    yield source

    @property
    def config_path(self) -> Path:
        """ Configuration path """
        if 'config_dir' in self._configuration_data:
            return Path(os.path.expanduser(self._configuration_data['config_dir']))
        if sys.platform.lower().startswith('win'):
            return Path(os.path.expanduser(os.path.join("~\\YTRSS")))
        return Path(os.path.expanduser(os.path.join("~/.config/ytrss")))

    @property
    def cache_path(self) -> Path:
        """ Path to cache directory """
        try:
            os.makedirs(self.config_path)
        except OSError:
            pass
        return Path(os.path.join(self.config_path, "cache"))

    @property
    def args(self) -> Sequence[str]:
        """ youtube_dl argument """
        if 'arguments' in self._configuration_data:
            if isinstance(self._configuration_data['arguments'], list):
                return self._configuration_data['arguments']
            return [self._configuration_data['arguments']]
        return []



from __future__ import unicode_literals

import unittest

from typing import List, Dict, Any


from ytrss.plugins.youtube_dl.movie import YouTubeMovie, InvalidDataMovieError

# This is tested class. Can have too many method
from ytrss.core.helpers.typing import Url


class TestMovie(unittest.TestCase):  # pylint: disable=R0904
    """
    Movie tests.
    """
    def test_input_argument_parse(self) -> None:
        """
        Testing usage of module.
        """
        test_suites: List[Dict[str, Any]] = [{
            "arg": "https://www.youtube.com/watch?v=I-JxpVFlaos",
            "identity": "ytdl:youtube:I-JxpVFlaos",
            "url": "https://www.youtube.com/watch?v=I-JxpVFlaos"
        }, {
            "arg": "http://www.youtube.com/watch?v=I-JxpVFlaos",
            "identity": "ytdl:youtube:I-JxpVFlaos",
            "url": "http://www.youtube.com/watch?v=I-JxpVFlaos"
        }, {
            "arg": "https://youtu.be/I-JxpVFlaos",
            "identity": "ytdl:youtube:I-JxpVFlaos",
            "url": "https://youtu.be/I-JxpVFlaos"
        }]

        for elem in test_suites:
            test_elem = YouTubeMovie(elem['arg'])
            self.assertEqual(test_elem.identity, elem['identity'])
            self.assertEqual(test_elem.url, elem['url'])

    def test_invalid_argument(self) -> None:
        """
        Testing invalid argument element.
        """
        test_suites: List[Url] = [
            Url("https://www.youtube.com/watch?v=fakeI-JxpVFlaosfs"),
            Url("https://www.youdupe.com/watch?v=I-JxpVFlaos"),
        ]
        for elem in test_suites:
            with self.assertRaises(InvalidDataMovieError):
                YouTubeMovie(elem)

    def test_comparation(self) -> None:
        """
        Testing comparation elements.
        """
        test_suites = [{
            "ob1": "https://www.youtube.com/watch?v=I-JxpVFlaos",
            "ob2": "https://www.youtube.com/watch?v=I-JxpVFlaos"
        }, {
            "ob1": "https://www.youtube.com/watch?v=I-JxpVFlaos",
            "ob2": "I-JxpVFlaos"
        }, {
            "ob1": "I-JxpVFlaos",
            "ob2": "I-JxpVFlaos"
        }, {
            "ob1": "I-JxpVFlaos",
            "ob2": "http://youtu.be/I-JxpVFlaos",
        }]
        for elem in test_suites:
            elem1 = YouTubeMovie(Url(elem['ob1']))
            elem2 = YouTubeMovie(Url(elem['ob2']))
            self.assertEqual(elem1,
                             elem2,
                             "Not equal [{} != {}".format(elem['ob1'],
                                                          elem['ob2']))



import abc

from typing import Any


class ConfigurationReader(metaclass=abc.ABCMeta):
    """ An abstract reader for configuration """

    @property
    @abc.abstractmethod
    def read_config(self) -> Any:
        """ Returns object with configuration data. """



import logging

from typing import Dict

_GREY = "\x1b[38;20m"
_BLUE = "\x1b[38;5;39m"
_YELLOW = "\x1b[33;20m"
_RED = "\x1b[31;20m"
_BOLD_RED = "\x1b[31;1m"
_RESET = "\x1b[0m"


class DebugFormatter(logging.Formatter):
    """
    Logging formatter for logs in debug mode
    """

    _message_schema = "[%(name)s] %(levelname)s - %(message)s (\"%(pathname)s:%(lineno)d\")"

    _FORMATS: Dict[int, str] = {
        logging.DEBUG: f"{_BLUE}{_message_schema}{_RESET}",
        logging.INFO: f"{_RESET}{_message_schema}{_RESET}",
        logging.WARNING: f"{_YELLOW}{_message_schema}{_RESET}",
        logging.ERROR: f"{_RED}{_message_schema}{_RESET}",
        logging.CRITICAL: f"{_BOLD_RED}{_message_schema}{_RESET}",
    }

    def format(self, record: logging.LogRecord) -> str:
        log_fmt = self._FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class ClientFormatter(logging.Formatter):
    """
    Logging formatter for logs in client app
    """

    _message_schema = "%(message)s"

    _FORMATS: Dict[int, str] = {
        logging.DEBUG: f"{_BLUE}{_message_schema}{_RESET}",
        logging.INFO: f"{_RESET}{_message_schema}{_RESET}",
        logging.WARNING: f"{_YELLOW}{_message_schema}{_RESET}",
        logging.ERROR: f"{_RED}{_message_schema}{_RESET}",
        logging.CRITICAL: f"{_BOLD_RED}{_message_schema}{_RESET}",
    }

    def format(self, record: logging.LogRecord) -> str:
        log_fmt = self._FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


logger = logging.getLogger("ytrss")



"""
Module with source data object
"""
from typing import Dict, Any

from ytrss.configuration.entity.destination_info import DestinationId


class Source:
    """
    Source data object

    This object contains an information about source, where the podcast can be downloaded
    """

    def __init__(self) -> None:
        self.url: str = "none"
        self.destination: DestinationId = DestinationId("")
        self.name: str = "<unknown>"
        self.enable: bool = True

    @property
    def json(self) -> Dict[str, Any]:
        """
        Dictionary representation of file
        """
        return {
            "name": self.name,
            "url": self.url,
            "destination": self.destination,
            "enable": self.enable
        }

    @staticmethod
    def from_json(json: Dict[str, Any]) -> 'Source':
        """ Create Source object from dictionary. """
        source = Source()
        if 'name' in json and isinstance(json['name'], str):
            source.name = json['name']
        if 'url' in json and isinstance(json['url'], str):
            source.url = json['url']
        if 'destination' in json and isinstance(json['destination'], str):
            source.destination = DestinationId(json['destination'])
        else:
            raise ValueError
        if 'enable' in json and isinstance(json['enable'], bool):
            source.enable = json['enable']
        return source



import os
import sys
from typing import Optional

from ytrss.configuration.configuration import Configuration, ConfigurationError, \
    ConfigurationFileNotExistsError
from ytrss.configuration.readers.base import ConfigurationReader
from ytrss.configuration.readers.default_configuration import DefaultConfiguration
from ytrss.configuration.readers.json_configuration import JsonConfiguration
from ytrss.configuration.readers.yaml_configuration import YamlConfiguration


class _ConfigurationFactory:
    """ Configuration builder """

    @classmethod
    def _create_from_file(
            cls,
            configuration_file: str
    ) -> ConfigurationReader:
        if not os.path.isfile(configuration_file):
            raise ConfigurationFileNotExistsError(configuration_file)
        if configuration_file.endswith(".json"):
            return JsonConfiguration(configuration_file)
        if configuration_file.endswith(".yml"):
            return YamlConfiguration(configuration_file)
        raise ConfigurationError("Not implements this type of file")

    @classmethod
    def _create_from_envs(cls, env_name: str) -> ConfigurationReader:
        conf_env_path = os.path.expanduser(os.environ[env_name])
        if os.path.isfile(os.path.join(conf_env_path, "config.yml")):
            return YamlConfiguration(os.path.join(conf_env_path, "config.yml"))
        if os.path.isfile(os.path.join(conf_env_path, "config.yml")):
            return YamlConfiguration(os.path.join(conf_env_path, "config.yml"))
        return DefaultConfiguration()

    @classmethod
    def _create_for_windows(cls, program_name: str) -> ConfigurationReader:
        if os.path.isfile(os.path.expanduser(f"~\\{program_name.upper()}\\config.yml")):
            return YamlConfiguration(f"~\\{program_name.upper()}\\config.yml")
        if os.path.isfile(os.path.expanduser(f"~\\{program_name.upper()}\\config.json")):
            return JsonConfiguration(f"~\\{program_name.upper()}\\config.json")
        return DefaultConfiguration()

    @classmethod
    def _create_for_unix(cls, program_name: str) -> ConfigurationReader:
        if os.path.isfile(os.path.expanduser(f"~/.config/{program_name}/config.yml")):
            return YamlConfiguration(f"~/.config/{program_name}/config.yml")
        if os.path.isfile(os.path.expanduser(f"~/.config/{program_name}/config.json")):
            return JsonConfiguration(f"~/.config/{program_name}/config.json")
        if os.path.isfile(f"/etc/{program_name}/config.yml"):
            return YamlConfiguration(f"/etc/{program_name}/config.yml")
        if os.path.isfile(f"/etc/{program_name}/config.json"):
            return JsonConfiguration(f"/etc/{program_name}/config.json")
        return DefaultConfiguration()

    @classmethod
    def _create_reader(
            cls,
            program_name: str,
            configuration_file: Optional[str]
    ) -> ConfigurationReader:
        """ Create Configuration Reader

        This method returns a configuration object.
        TODO: Write about localization of configuration files.
        """

        if configuration_file is not None and configuration_file != "":
            return cls._create_from_file(configuration_file)
        if "TOX_ENV_DIR" in os.environ:
            return cls._create_from_envs("TOX_ENV_DIR")
        if f"PYTHON_{program_name.upper()}_CONFIG" in os.environ:
            return cls._create_from_file(f"PYTHON_{program_name.upper()}_CONFIG")
        if sys.platform.lower().startswith('win'):
            return cls._create_for_windows(program_name)
        return cls._create_for_unix(program_name)

    def __call__(
            self,
            program_name: str = "python",
            configuration_file: Optional[str] = None
    ) -> Configuration:
        return Configuration(self._create_reader(program_name, configuration_file))


create_configuration = _ConfigurationFactory()



from argparse import Namespace

from ytrss.commands import BaseCommand
from ytrss.core.helpers.logging import logger


class ConfigurationCommand(BaseCommand):
    """
    Print configuration
    """

    def __init__(self) -> None:
        BaseCommand.__init__(self, "configuration")

    def run(self, options: Namespace) -> int:
        logger.info("subscriptions:")
        for source in self.manager_service.configuration.sources:
            logger.info("\t%s [%s]%s (%s)",
                        source.name,
                        source.url,
                        " [disabled]" if not source.enable else "",
                        source.destination)

        logger.info("\ndestinations:")
        for destination in self.manager_service.destination_manager.destinations:
            logger.info("\t[%s] => %s", destination.identity, destination.info.title)

        logger.info("\nplugins:")
        for plugin in self.manager_service.plugin_manager.plugins:
            logger.info("\t[%s] => %s", plugin.plugin_name, plugin.plugin_description)
        return 0



import abc
from typing import Iterator, Sequence

from ytrss.configuration.entity.destination_info import DestinationId, DestinationInfo
from ytrss.core.entity.downloaded_movie import DownloadedMovie

from ytrss.core.helpers.typing import Path


class DestinationError(Exception):
    """ Destination error

    The error raised when Destination object cannot be created
    """


class Destination(metaclass=abc.ABCMeta):
    """ Destination object

    This object represents a place where the file should be saved after download.
    """

    @property
    @abc.abstractmethod
    def identity(self) -> DestinationId:
        """ Identity of destination

        A string representation of object, that can be defined in json file by user.
        """

    @property
    @abc.abstractmethod
    def info(self) -> DestinationInfo:
        """ Information about destination. """

    @property
    @abc.abstractmethod
    def saved_movies(self) -> Iterator[DownloadedMovie]:
        """ A list of movies saved in destination. """

    def on_finish(self) -> None:
        """ An optional method that generate additional files, or make an additional
         things after download like sending to external service. """

    @abc.abstractmethod
    def save(self, files: Sequence[Path]) -> None:
        """ Save file from downloader cache to destination. """



import os
from typing import Any
import yaml
from ytrss.configuration.configuration import ConfigurationError, ConfigurationFileNotExistsError
from ytrss.configuration.readers.base import ConfigurationReader


class YamlConfigurationParseError(ConfigurationError):
    """ Yaml configuration error """


class YamlConfigurationFileNotExistsError(ConfigurationFileNotExistsError):
    """ Yaml configuration file not exists error. """


class YamlConfiguration(ConfigurationReader):
    """ Configuration yaml object """

    @property
    def read_config(self) -> Any:
        return self.__configuration

    def __init__(self, conf_file: str) -> None:
        if not os.path.isfile(os.path.expanduser(conf_file)):
            raise ConfigurationError("File not exists")
        try:
            with open(os.path.expanduser(conf_file)) as data_file:
                # pylint: disable=E1120
                data = yaml.safe_load(data_file)
        except FileNotFoundError:
            raise YamlConfigurationFileNotExistsError()

        if data == {}:
            raise YamlConfigurationParseError("Cannot find data from file")

        self.__configuration = data



import unittest

from ytrss.configuration.entity.source import Source
from ytrss.plugins.youtube.youtube_channel_source_downloader import YouTubeChannelSourceDownloader
from ytrss.plugins.youtube.youtube_named_channel_source_downloader import \
    YouTubeNamedChannelSourceDownloader
from ytrss.plugins.youtube.youtube_playlist_source_downloader import YouTubePlaylistSourceDownloader


class TestFinder(unittest.TestCase):  # pylint: disable=R0904
    """ Test of finder """

    def test_channel_with_code_find(self) -> None:
        """ Test channel finder with code. """
        source = YouTubeChannelSourceDownloader(Source.from_json(dict(
            url="https://www.youtube.com/channel/UC5SdvjkYXdGAqmqAGlxTnpw",
            destination="default",
        )))
        movies = list(source.movies)
        self.assertGreater(len(movies), 0)
        for movie, _ in movies:
            print(f"url: {movie}")
            self.assertTrue(movie.url is not None
                            and movie.url.startswith("https://www.youtube.com/watch?v="))

    def test_channel_with_name_find(self) -> None:
        """ Test channel finder with name. """
        source = YouTubeNamedChannelSourceDownloader(Source.from_json(dict(
            url="https://www.youtube.com/c/MinnivaOfficial",
            destination="default",
        )))
        movies = list(source.movies)
        self.assertGreater(len(movies), 0)
        for movie, _ in movies:
            print(f"url: {movie}")
            self.assertTrue(movie.url is not None
                            and movie.url.startswith("https://www.youtube.com/watch?v="))

    def test_playlist_find(self) -> None:
        """ Test playlist finder. """
        source = YouTubePlaylistSourceDownloader(Source.from_json(dict(
            url="https://www.youtube.com/playlist?list=PL6ZLc-zZUnxlkB9t8CcpFZeV6V5I_cVgu",
            destination="default",
        )))
        movies = list(source.movies)
        self.assertGreater(len(movies), 0)
        for movie, _ in movies:
            print(f"url: {movie}")
            self.assertTrue(movie.url is not None
                            and movie.url.startswith("https://www.youtube.com/watch?v="))



"""
example test module.
"""

import unittest


class TestExample(unittest.TestCase):
    """
    Example test object
    """

    def test_example(self) -> None:
        """
        test example
        """
        self.assertTrue(True)  # pylint: disable=W1503


if __name__ == "__main__":
    unittest.main()



from typing import List, Sequence

from ytrss.configuration.entity.configuration_data import YtrssConfiguration
from ytrss.configuration.entity.destination_info import DestinationInfo
from ytrss.configuration.entity.source import Source
from ytrss.core.entity.destination import Destination, DestinationError
from ytrss.core.entity.downloader import Downloader, DownloaderError
from ytrss.core.entity.movie import Movie, MovieError
from ytrss.core.entity.plugin import BasePlugin
from ytrss.core.entity.source_downloader import SourceDownloader, SourceDownloaderError
from ytrss.core.helpers.typing import Url
from ytrss.plugins.default.plugin import DefaultPlugin
from ytrss.plugins.rss.plugin import RssPlugin
from ytrss.plugins.youtube.plugin import YouTubePlugin
from ytrss.plugins.youtube_dl.plugin import YouTubeDlPlugin


class PluginManager(BasePlugin):
    """ Plugin Manager """

    def __init__(self) -> None:
        self._plugins: List[BasePlugin] = [
            DefaultPlugin(),
            RssPlugin(),
            YouTubePlugin(),
            YouTubeDlPlugin()
        ]

    @property
    def plugins(self) -> Sequence[BasePlugin]:
        """ List of plugins"""
        return self._plugins

    def create_movie(self, url: Url) -> Movie:
        for plugin in self._plugins:
            try:
                movie = plugin.create_movie(url)  # pylint: disable=E1128
                if movie is not None:
                    return movie
            except Exception:  # pylint: disable=W0703
                pass
        raise MovieError()

    def create_destination(self, info: DestinationInfo) -> Destination:
        for plugin in self._plugins:
            try:
                destination = plugin.create_destination(info)
                if destination is not None:
                    return destination
            except Exception:  # pylint: disable=W0703
                pass
        raise DestinationError()

    def create_downloader(self, movie: Movie, configuration: YtrssConfiguration) -> Downloader:
        for plugin in self._plugins:
            try:
                downloader = plugin.create_downloader(movie, configuration)  # pylint: disable=E1128
                if downloader is not None:
                    return downloader
            except Exception:  # pylint: disable=W0703
                pass
        raise DownloaderError()

    def create_source_downloader(self, source: Source) -> SourceDownloader:
        for plugin in self._plugins:
            try:
                source_downloader = plugin.create_source_downloader(source)  # pylint: disable=E1128
                if source_downloader is not None:
                    return source_downloader
            except Exception:  # pylint: disable=W0703
                pass
        raise SourceDownloaderError()



import abc
from argparse import ArgumentParser, Namespace

from ytrss.core.managers.manager_service import default_manager_service


class BaseCommand(metaclass=abc.ABCMeta):
    """ Abstract class represents a command in terminal client """

    def __init__(self, name: str) -> None:
        self.manager_service = default_manager_service()
        self.name = name

    @abc.abstractmethod
    def run(self, options: Namespace) -> int:
        """ A main work that be done in subcommand from terminal client """

    def arg_parser(self, parser: ArgumentParser) -> None:
        """ Parse argument

        In this method terminal arguments from subcommand should be registered in main subparser.
        """

    def __call__(self, options: Namespace) -> int:
        return self.run(options)



import abc
from typing import Iterable, Tuple

from ytrss.configuration.entity.destination_info import DestinationId
from ytrss.core.entity.movie import Movie


class SourceDownloaderError(Exception):
    """ This is source downloader error """


class SourceDownloader(metaclass=abc.ABCMeta):
    """ Source downloader abstract class.

    This class represents a source of movies. It should check a source and returns an iterable
    list of movies available to download.
    """

    @property
    @abc.abstractmethod
    def movies(self) -> Iterable[Tuple[Movie, DestinationId]]:
        """ A list of movies available to download with destination id of this movie"""



from typing import Optional

from ytrss.core.entity.downloader import Downloader

from ytrss.configuration.entity.configuration_data import YtrssConfiguration
from ytrss.core.entity.movie import Movie, MovieError
from ytrss.core.entity.plugin import Plugin
from ytrss.core.helpers.typing import Url
from ytrss.plugins.youtube_dl.movie import YouTubeMovie
from ytrss.plugins.youtube_dl.youtube_downloader import YouTubeDownloader


class YouTubeDlPlugin(Plugin):
    """ Plugin implements mechanisms from youtube_dl """

    def create_movie(self, url: Url) -> Optional[Movie]:
        try:
            return YouTubeMovie(url)
        except MovieError:
            return None

    def create_downloader(self, movie: Movie, configuration: YtrssConfiguration) -> Optional[Downloader]:
        return YouTubeDownloader(movie, configuration)



import os.path
from typing import List
from jinja2 import Environment, FileSystemLoader

import ytrss
from ytrss.configuration.entity.destination_info import DestinationInfo
from ytrss.core.entity.downloaded_movie import DownloadedMovie
from ytrss.plugins.rss.podcast.helpers import format_str, format_desc, format_date


class Podcast:
    """
    Class that generate rss file from DownloaderMovie
    """

    def __init__(self, podcast_info: DestinationInfo) -> None:
        self.__podcast_info = podcast_info
        self.__items: List[DownloadedMovie] = []

    def add_item(self, movie: DownloadedMovie) -> None:
        """ Add downloaded movie to rss """
        self.__items.append(movie)

    def generate(self) -> str:
        """ Generate rss file for object """
        env = Environment(
            loader=FileSystemLoader(os.path.join(ytrss.__path__[0], "templates")),  # type: ignore
        )
        env.filters["format_str"] = format_str
        env.filters["format_desc"] = format_desc
        env.filters["format_date"] = format_date

        template = env.get_template("podcast.xml")
        return template.render(podcast_info=self.__podcast_info, movies=self.__items)



import os
import shutil
from argparse import Namespace, ArgumentParser

from ytrss.core.entity.downloader import DownloaderError

from ytrss.commands import BaseCommand
from ytrss.core.entity.movie import MovieError
from ytrss.core.helpers.logging import logger
from ytrss.core.helpers.typing import Url


class DownloadCommand(BaseCommand):
    """
    Immediately download a movie to current location.
    """

    def __init__(self) -> None:
        BaseCommand.__init__(self, "download")

    def arg_parser(self, parser: ArgumentParser) -> None:
        parser.add_argument("url",
                            help="Url that should be downloaded")

    def run(self, options: Namespace) -> int:
        try:
            movie = self.manager_service.plugin_manager.create_movie(Url(options.url))

        except MovieError:
            logger.error("This is not valid url or movie not exists: %s", options.url)
            return 1

        try:
            downloaded_movie = self\
                .manager_service\
                .plugin_manager\
                .create_downloader(movie,
                                   self.manager_service.configuration).download()
        except DownloaderError:
            logger.error("There are no downloaded files for movie from: %s", movie.url)
            return 1

        destination_path = os.path.join(os.getcwd(), movie.identity)
        os.makedirs(destination_path, exist_ok=True)
        for file in downloaded_movie.data_paths:
            try:
                shutil.move(file, destination_path)
            except shutil.Error as ex:
                logger.error(ex)
        return 0



import abc

from ytrss.core.entity.downloaded_movie import DownloadedMovie


class DownloaderError(Exception):
    """
    Error in Downloaders.
    """


class Downloader(metaclass=abc.ABCMeta):
    """ An abstract class of downloader

    The class represents an object that can download movie. It should return a sequence of paths to
    objects that has an information about files.
    """

    @abc.abstractmethod
    def download(self) -> DownloadedMovie:
        """
        Download movie

        The method which should download movie and return a DownloadedMovie object.
        """



import re

from ytrss.configuration.entity.source import Source
from ytrss.core.entity.source_downloader import SourceDownloaderError
from ytrss.core.helpers.typing import Url
from ytrss.plugins.youtube.base_youtube_source_downloader import BaseYouTubeSourceDownloader


class YouTubePlaylistSourceDownloader(BaseYouTubeSourceDownloader):
    """ YouTube playlist source Downloader

    Implementation of source downloader for playlist on YouTube
    """

    def __init__(self, source: Source) -> None:
        match = re.match(r"https?:\/\/(www\.)?youtube\.com\/playlist\?.*list=(?P<code>[\dA-Za-z_\-]+)", source.url)
        if match is None:
            raise SourceDownloaderError()
        self._code = match.group('code')
        BaseYouTubeSourceDownloader.__init__(self, source)

    @property
    def source_url(self) -> Url:
        return Url(f"https://www.youtube.com/feeds/videos.xml?playlist_id={self._code}")



import os
import sqlite3
from contextlib import contextmanager
from sqlite3 import Connection
from typing import Iterable, Iterator, Tuple

from ytrss.configuration.entity.configuration_data import YtrssConfiguration

from ytrss.configuration.entity.destination_info import DestinationId
from ytrss.core.entity.movie import Movie, MovieError
from ytrss.core.helpers.logging import logger
from ytrss.core.managers.plugin_manager import PluginManager

from ytrss.database.database import Database, DatabaseStatus


class SqliteDatabase(Database):
    """
    Database implementation in sqlite3
    """

    @contextmanager
    def _database_handler(self) -> Iterator[Connection]:
        conn = sqlite3.connect(self._db_file)
        yield conn
        conn.close()

    def __init__(self, configuration: YtrssConfiguration, plugin_manager: PluginManager):
        self.__configuration = configuration
        self._plugin_manager = plugin_manager
        self._db_file = os.path.join(configuration.config_path, "ytrss.db")
        with self._database_handler() as conn:
            conn.execute('''
                CREATE TABLE IF NOT EXISTS movies_tasks (
                    identity TEXT NOT NULL UNIQUE,
                    url TEXT NOT NULL,
                    destination TEXT NOT NULL,
                    created TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    try INT DEFAULT 0,
                    state CHAR(10) DEFAULT "wait",
                    comment TEXT
                )
            ''')

    def change_type(self, movie: Movie, db_type: DatabaseStatus) -> None:
        with self._database_handler() as conn:
            conn.execute(f'''
                UPDATE movies_tasks
                set state = "{db_type.value}"
                where identity = "{movie.identity}"
            ''')
            conn.commit()

    def is_new(self, movie: Movie, destination: DestinationId) -> bool:
        with self._database_handler() as conn:
            cursor = conn.execute(f'''
                SELECT identity
                FROM movies_tasks
                WHERE identity = "{movie.identity}" and state IN ("wait", "error", "progress")
            ''')
            rows = cursor.fetchall()
        return len(rows) > 0

    def movies(self) -> Iterable[Tuple[Movie, DestinationId]]:
        with self._database_handler() as conn:
            cursor = conn.execute('''
                SELECT identity, url, destination
                FROM movies_tasks
                WHERE state IN ("wait", "error", "progress")
            ''')
            rows = cursor.fetchall()

        for row in rows:
            try:
                movie = self._plugin_manager.create_movie(row[1])
            except MovieError:
                logger.error("Not valid url or movie not exists: %s", row[1])
                continue
            destination = DestinationId(row[2])
            if self.is_new(movie, destination):
                yield movie, destination

    def queue_mp3(self, movie: Movie, destination: DestinationId) -> bool:
        """
        Add address to download.
        """
        try:
            with self._database_handler() as conn:
                conn.execute(f'''
                    INSERT INTO movies_tasks (identity, url, destination)
                    VALUES ('{movie.identity}', '{movie.url}', '{destination}')
                ''')
                conn.commit()
        except sqlite3.IntegrityError:
            return False
        return True



#!/usr/bin/env python3
from __future__ import unicode_literals
from setuptools import setup
from setuptools import find_packages


def read_description(local_current_package):
    module_doc = local_current_package.__doc__.splitlines()
    result = ""
    for line in module_doc:
        if line:
            result = line
            break
    return result


current_package = __import__('ytrss')

data_files = []

setup(
    name=current_package.__title__,
    version=current_package.__version__,
    license=current_package.__license__,
    author=current_package.__author__,
    author_email=current_package.__author_email__,
    description=read_description(current_package),
    python_requires='>={}.{}'.format(*current_package.__required_python__),
    long_description=open("README.rst").read(),
    url=current_package.__url__,
    project_urls={
        "Source": "https://github.com/rafyco/ytrss",
        "Tracker": "https://github.com/rafyco/ytrss/issues",
        "Changelog": "https://github.com/rafyco/ytrss/blob/master/dosc/changelog.rst",
        "Documentation": "https://ytrss.readthedocs.io"
    },
    packages=find_packages(),
    include_package_data=True,
    package_dir={'ytrss': 'ytrss'},
    classifiers=[
        'Environment :: Console',
        'Development Status :: 3 - Alpha',
        'Topic :: Multimedia :: Sound/Audio',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ],
    install_requires=[
        'pyyaml',
        'astroid ~= 2.5',
        'jinja2 ~= 3.0.3',
        'youtube_dl ~= 2021.12.17',
        'locks ~= 0.1.1'
    ],
    extras_require={
        "optional": [
            "argcomplete"
        ],
        "unittests": [
            "pytest"
        ],
        "style": [
            "pylint == 2.5.3",
            "pep8 == 1.7.1",
            "pycodestyle == 2.8.0"
        ],
        "typing": [
            "mypy == 0.910",
            "types-PyYAML == 6.0.8"
        ],
        "documentation": [
            "Sphinx == 4.3.1",
            "sphinx-epytext == 0.0.4",
            "sphinx-autorun == 1.1.1"
        ]
    },
    entry_points={
        'console_scripts': [
            'ytrss = ytrss.client:main',
        ]
    },
    command_options={
        'build_spninx': {
            'project': ('setup.py', current_package.__title__),
            'version': ('setup.py', current_package.__version__),
            'release': ('setup.py', current_package.__version__),
            'source_dir': ('setup.py', 'docs')
        }
    },
    platforms="Any",
    keywords="youtube, console, download, rss, mp3, service"
)



import os
from typing import Optional, Dict, Any, NewType
from ytrss.core.helpers.typing import Path

DestinationId = NewType("DestinationId", str)


class DestinationInfo:
    """ Destination info

    An object with destination information from json object.
    """

    def __init__(self, identity: DestinationId) -> None:
        self.identity = identity
        self.type: str = ""
        self.title: str = "unknown title"
        self.author: str = "Nobody"
        self.language: str = "pl-pl"
        self.link: str = "http://youtube.com"
        self.desc: str = "No description"
        self.url_prefix: str = ""
        self.img: Optional[str] = None
        self.output_path: Optional[Path] = None

    @staticmethod
    def from_json(json: Optional[Dict[str, Any]], identity: DestinationId) -> 'DestinationInfo':
        """ Create object from json dictionary. """
        podcast = DestinationInfo(identity)
        if json is not None:
            if 'title' in json and isinstance(json['title'], str):
                podcast.title = json['title']
            if 'type' in json and isinstance(json['type'], str):
                podcast.type = json['type']
            if 'author' in json and isinstance(json['author'], str):
                podcast.author = json['author']
            if 'language' in json and isinstance(json['language'], str):
                podcast.language = json['language']
            if 'link' in json and isinstance(json['link'], str):
                podcast.link = json['link']
            if 'desc' in json and isinstance(json['desc'], str):
                podcast.desc = json['desc']
            if 'url_prefix' in json and isinstance(json['url_prefix'], str):
                podcast.url_prefix = json['url_prefix']
            if 'img' in json and isinstance(json['img'], str):
                podcast.img = json['img']
            if 'path' in json and isinstance(json['path'], str):
                podcast.output_path = Path(os.path.expanduser(json['path']))
        return podcast



import abc

from typing import Optional

from ytrss.core.entity.downloader import Downloader

from ytrss.configuration.entity.configuration_data import YtrssConfiguration
from ytrss.configuration.entity.source import Source
from ytrss.core.entity.destination import Destination

from ytrss.configuration.entity.destination_info import DestinationInfo

from ytrss.core.entity.movie import Movie
from ytrss.core.entity.source_downloader import SourceDownloader
from ytrss.core.helpers.string_utils import first_line

from ytrss.core.helpers.typing import Url


class BasePlugin(metaclass=abc.ABCMeta):
    """ Plugin's interface """

    @property
    def plugin_name(self) -> str:
        """ Plugin's name """
        return type(self).__name__

    @property
    def plugin_description(self) -> str:
        """ Short description of plugin """
        short_desc = first_line(self.__doc__)
        return short_desc if short_desc != "" else f"{self.plugin_name}: not description"

    @abc.abstractmethod
    def create_movie(self, url: Url) -> Optional[Movie]:
        """ Method that creates a Movie object."""

    @abc.abstractmethod
    def create_destination(self, info: DestinationInfo) -> Optional[Destination]:
        """ Method that creates destination object. """

    @abc.abstractmethod
    def create_downloader(self, movie: Movie, configuration: YtrssConfiguration) -> Optional[Downloader]:
        """ Method that creates a downloader object. """

    @abc.abstractmethod
    def create_source_downloader(self, source: Source) -> Optional[SourceDownloader]:
        """ Method that creates a source downloader object. """


class Plugin(BasePlugin):
    """ An implementation of plugin that not return any of new objects.

    This object can be used to implement new plugin object. All method that creates objects, return None.
    """

    def create_destination(self, info: DestinationInfo) -> Optional[Destination]:
        return None

    def create_downloader(self, movie: Movie, configuration: YtrssConfiguration) -> Optional[Downloader]:
        return None

    def create_source_downloader(self, source: Source) -> Optional[SourceDownloader]:
        return None

    def create_movie(self, url: Url) -> Optional[Movie]:
        return None



from abc import abstractmethod

from typing import Iterable, Tuple
from urllib.request import urlopen
from xml.dom import minidom

from ytrss.configuration.entity.destination_info import DestinationId
from ytrss.configuration.entity.source import Source
from ytrss.core.entity.movie import Movie
from ytrss.core.entity.source_downloader import SourceDownloader
from ytrss.core.helpers.logging import logger
from ytrss.core.helpers.typing import Url
from ytrss.plugins.youtube_dl.movie import YouTubeMovie


class BaseYouTubeSourceDownloader(SourceDownloader):
    """ Base YouTube source downloader

    Object that implements source downloader that look for movie from YouTube.
    """

    def __init__(self, source: Source) -> None:
        self.url = source.url
        self.destination = source.destination

    @property
    @abstractmethod
    def source_url(self) -> Url:
        """ Rss url with movies on YouTube """

    @property
    def movies(self) -> Iterable[Tuple[Movie, DestinationId]]:
        logger.debug("Source url: %s", self.url)

        try:
            xml_str = urlopen(str(self.source_url)).read()
            xmldoc = minidom.parseString(xml_str)
            tags = xmldoc.getElementsByTagName('link')
        # We want catch every exception in ulr like invalid channel or web
        except Exception:  # pylint: disable=W0703
            logger.error("Problem with url: %s", self.url)
            return
        for elem in tags:
            url: Url = Url(elem.getAttribute("href"))
            if "watch?v=" in url:  # pylint: disable=E1135
                try:
                    yield YouTubeMovie(url), self.destination
                except Exception:  # pylint: disable=W0703
                    logger.error("Problem with url in source: %s", self.url)
